(***********************************************************************)
(*                                                                     *)
(*                               Options                               *)
(*                                                                     *)
(*             projet Estime & Cristal, INRIA-Rocquencourt             *)
(*                                                                     *)
(*  Copyright 2005-2006 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.mli,v 1.1 2007/01/17 23:36:00 fclement Exp $ *)

(** To add options to the command line of the program.
   We consider three kinds of options: regular, debug, and report options.
   Report options are used to collect information during the execution of the
   program (e.g. for the monitoring of iterative algorithms).
   Debug options are used to trace the execution when a program does not behave
   properly. *)

type option_name = string;;
type manual_text = string;;

val all : unit -> (option_name * Arg.spec * manual_text) list;;
(** Return the entire list of options that have ever been defined by one of the
   option creation functions of this module.
   In addition this list contains the two built-in options [-d] and [--debug]
   that fire all the debug options defined so far. *)

val add : option_name -> Arg.spec -> manual_text -> unit;;
(** [add oname spec man] add the option [oname] to the command line
   with specification [spec] and manual info [man]. *)

val add_all : (option_name * Arg.spec * manual_text) list -> unit;;
(** [add_all options] adds all the options of the list [options]. *)

val add_debug : option_name -> Arg.spec -> manual_text -> unit;;
(** [add_debug oname spec man] add the debug option [oname] to the command line
   with specification [spec] and manual info [man].
   Built-in options [-d] and [--debug] will automatically fire this debug
   option. *)

val add_all_debug : (option_name * Arg.spec * manual_text) list -> unit;;
(** [add_all_debug debug_options] adds all the debug options of the list
   [debug_options]. *)

val add_report : option_name -> Arg.spec -> manual_text -> unit;;
(** [add_report oname spec man] add the report option [oname] to the command
   line with specification [spec] and manual info [man].
   Built-in options [-r] and [--report] will automatically fire this report
   option. *)

val add_all_report : (option_name * Arg.spec * manual_text) list -> unit;;
(** [add_all_report report_options] adds all the report options of the list
   [report_options]. *)

val debug : option_name -> manual_text -> string -> unit;;
(** [debug oname man] add the option flag [oname] to the command line option
   list with associated manual info [man].
   The flag is false by default and can be set using the command line option
   [oname].
   [debug] returns a function that prints its argument on [stderr] 
   if the flag is set. *)

val flag : bool -> option_name -> manual_text -> bool ref;;
(** [flag init oname man] add the option flag [oname] to the command line option
   list with associated manual info [man] and returns the corresponding boolean
   reference flag. This reference flag has initial boolean value [init],
   and is modified with the command line option [oname]: if the initial
   value [init] is [false] then the option [oname] sets the flag;
   otherwise the option [oname] clears the flag. *)
